@@ -13,15 +13,17 @@
 package Directory::Queue::Normal;
 use strict;
 use warnings;
-our $VERSION  = "1.7_2";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.11 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.10";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.14 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
 #
 
-use Directory::Queue qw(_name /Regexp/ /file/ /special/);
+use Directory::Queue qw(_create _name _touch /Regexp/ /special/);
+use Encode qw(encode decode FB_CROAK LEAVE_SRC);
 use No::Worries::Die qw(dief);
+use No::Worries::File qw(file_read file_write);
 use No::Worries::Stat qw(ST_MTIME ST_NLINK);
 use No::Worries::Warn qw(warnf);
 use POSIX qw(:errno_h);
@@ -205,7 +207,7 @@ sub new : method {
 
     # default object
     $self = __PACKAGE__->SUPER::_new(%option);
-    foreach my $name (qw(path umask)) {
+    foreach my $name (qw(path maxlock maxtemp rndhex umask)) {
         delete($option{$name});
     }
     # default options
@@ -246,7 +248,7 @@ sub new : method {
     }
     # create directories
     foreach my $name (TEMPORARY_DIRECTORY, OBSOLETE_DIRECTORY) {
-        $path = $self->{path} . "/" . $name;
+        $path = $self->{path}."/".$name;
         _special_mkdir($path, $self->{umask}) unless -d $path;
     }
     # so far so good...
@@ -268,7 +270,7 @@ sub count : method {
     }
     # count sub-directories
     foreach my $name (@list) {
-        $subdirs = $_SubDirs->($self->{path} . "/" . $name);
+        $subdirs = $_SubDirs->($self->{path}."/".$name);
         $count += $subdirs if $subdirs;
     }
     # that's all
@@ -285,8 +287,8 @@ sub _is_locked ($$;$) {
     my($self, $name, $time) = @_;
     my($path, @stat);
 
-    $path = $self->{path} . "/" . $name;
-    return(0) unless -d $path . "/" . LOCKED_DIRECTORY;
+    $path = $self->{path}."/".$name;
+    return(0) unless -d $path."/".LOCKED_DIRECTORY;
     return(1) unless defined($time);
     @stat = lstat($path);
     unless (@stat) {
@@ -319,7 +321,7 @@ sub lock : method { ## no critic 'ProhibitBuiltinHomonyms'
 
     _check_element($element);
     $permissive = 1 unless defined($permissive);
-    $path = $self->{path} . "/" . $element . "/" . LOCKED_DIRECTORY;
+    $path = $self->{path}."/".$element."/".LOCKED_DIRECTORY;
     if (defined($self->{umask})) {
         $oldumask = umask($self->{umask});
         $success = mkdir($path);
@@ -337,7 +339,7 @@ sub lock : method { ## no critic 'ProhibitBuiltinHomonyms'
         # otherwise this is unexpected...
         dief("cannot mkdir(%s): %s", $path, $!);
     }
-    $path = $self->{path} . "/" . $element;
+    $path = $self->{path}."/".$element;
     unless (lstat($path)) {
         if ($permissive and $! == ENOENT) {
             # RACE: the element directory does not exist anymore
@@ -371,7 +373,7 @@ sub unlock : method {
     my($path);
 
     _check_element($element);
-    $path = $self->{path} . "/" . $element . "/" . LOCKED_DIRECTORY;
+    $path = $self->{path}."/".$element."/".LOCKED_DIRECTORY;
     unless (rmdir($path)) {
         if ($permissive) {
             # RACE: the element directory or its lock does not exist anymore
@@ -384,6 +386,16 @@ sub unlock : method {
     return(1);
 }
 
+#
+# touch an element to indicate that it is still being used
+#
+
+sub touch : method {
+    my($self, $element) = @_;
+
+    _touch($self->{"path"}."/".$element);
+}
+
 #
 # remove a locked element from the queue
 #
@@ -396,11 +408,11 @@ sub remove : method {
     dief("cannot remove %s: not locked", $element)
         unless _is_locked($self, $element);
     # move the element out of its intermediate directory
-    $path = $self->{path} . "/" . $element;
+    $path = $self->{path}."/".$element;
     while (1) {
         $temp = $self->{path}
-            . "/" . OBSOLETE_DIRECTORY
-            . "/" . _name($self->{rndhex});
+           ."/".OBSOLETE_DIRECTORY
+           ."/"._name($self->{rndhex});
         rename($path, $temp) and last;
         dief("cannot rename(%s, %s): %s", $path, $temp, $!)
             unless $! == ENOTEMPTY or $! == EEXIST;
@@ -410,7 +422,7 @@ sub remove : method {
     foreach my $name (_special_getdir($temp, "strict")) {
         next if $name eq LOCKED_DIRECTORY;
         if ($name =~ /^($_FileRegexp)$/o) {
-            $path = $temp . "/" . $1; # untaint
+            $path = $temp."/".$1; # untaint
         } else {
             dief("unexpected file in %s: %s", $temp, $name);
         }
@@ -418,7 +430,7 @@ sub remove : method {
         dief("cannot unlink(%s): %s", $path, $!);
     }
     # remove the locked directory
-    $path = $temp . "/" . LOCKED_DIRECTORY;
+    $path = $temp."/".LOCKED_DIRECTORY;
     while (1) {
         rmdir($path) or dief("cannot rmdir(%s): %s", $path, $!);
         rmdir($temp) and return;
@@ -430,6 +442,36 @@ sub remove : method {
     }
 }
 
+#
+# read a binary file and return a reference to the corresponding data
+#
+
+sub _file_read_bin ($) {
+    my($path) = @_;
+    my($data);
+
+    file_read($path, data => \$data);
+    return(\$data);
+}
+
+#
+# read a UTF-8 encoded file and return a reference to the corresponding string
+#
+
+sub _file_read_utf8 ($) {
+    my($path) = @_;
+    my($data, $string);
+
+    file_read($path, data => \$data);
+    eval {
+        local $SIG{__WARN__} = sub { die($_[0]) };
+        $string = decode("UTF-8", $data, FB_CROAK);
+    };
+    return(\$string) unless $@;
+    $@ =~ s/\s+at\s.+?\sline\s+\d+\.?$//;
+    dief("cannot UTF-8 decode %s: %s", $path, $@);
+}
+
 #
 # get an element from a locked element
 #
@@ -487,13 +529,13 @@ sub _insertion_directory ($) {
     # handle the case with no directories yet
     unless (@list) {
         $new = sprintf("%08x", 0);
-        _special_mkdir($self->{path} . "/" . $new, $self->{umask});
+        _special_mkdir($self->{path}."/".$new, $self->{umask});
         return($new);
     }
     # check the last directory
     @list = sort(@list);
     $new = pop(@list);
-    $subdirs = $_SubDirs->($self->{path} . "/" . $new);
+    $subdirs = $_SubDirs->($self->{path}."/".$new);
     if (defined($subdirs)) {
         return($new) if $subdirs < $self->{maxelts};
         # this last directory is now full... create a new one
@@ -504,7 +546,7 @@ sub _insertion_directory ($) {
     }
     # we need a new directory
     $new = sprintf("%08x", hex($new) + 1);
-    _special_mkdir($self->{path} . "/" . $new, $self->{umask});
+    _special_mkdir($self->{path}."/".$new, $self->{umask});
     return($new);
 }
 
@@ -513,8 +555,8 @@ sub _insertion_directory ($) {
 #
 
 sub _add_data ($$$) {
-    my($self, $data, $temp) = @_;
-    my($ref, $utf8);
+    my($self, $data, $tempdir) = @_;
+    my($ref, $utf8, $tmp, $path, $fh);
 
     foreach my $name (keys(%{ $data })) {
         dief("unexpected data: %s", $name) unless $self->{type}{$name};
@@ -541,17 +583,19 @@ sub _add_data ($$$) {
             dief("unexpected data type in %s: %s",
                  $name, $self->{type}{$name});
         }
-        eval {
-            _file_write("$temp/$name", $utf8, $self->{umask}, $ref);
-        };
-        if ($@) {
-            if ($@ =~ /^Wide character in /) {
-                dief("unexpected wide character in %s: %s",
-                     $name, $data->{$name});
-            } else {
-                die($@);
+        if ($utf8) {
+            eval {
+                $tmp = encode("UTF-8", ${ $ref }, FB_CROAK|LEAVE_SRC);
+            };
+            if ($@) {
+                $@ =~ s/\s+at\s.+?\sline\s+\d+\.?$//;
+                dief("unexpected character in %s: %s", $name, $@);
             }
+            $ref = \$tmp;
         }
+        $path = "$tempdir/$name";
+        $fh = _create($path, $self->{umask}, "strict");
+        file_write($path, handle => $fh, data => $ref);
     }
 }
 
@@ -562,13 +606,11 @@ sub _add_data ($$$) {
 #  - the destination directory must _not_ be created beforehand as it would
 #    be seen as a valid (but empty) element directory by an other process,
 #    we therefor use rename() from a temporary directory
-#  - syswrite() used in _file_write() may die with a "Wide character"
-#    "severe warning", we trap it here to provide better information
 #
 
 sub add : method {
     my($self, @data) = @_;
-    my($data, $temp, $dir, $new, $path, $ref, $utf8);
+    my($data, $tempdir, $dir, $new, $path, $ref, $utf8);
 
     dief("unknown schema") unless $self->{type};
     if (@data == 1) {
@@ -581,18 +623,18 @@ sub add : method {
             unless defined($data->{$name});
     }
     while (1) {
-        $temp = $self->{path}
-            . "/" . TEMPORARY_DIRECTORY
-            . "/" . _name($self->{rndhex});
-        last if _special_mkdir($temp, $self->{umask});
+        $tempdir = $self->{path}
+           ."/".TEMPORARY_DIRECTORY
+           ."/"._name($self->{rndhex});
+        last if _special_mkdir($tempdir, $self->{umask});
     }
-    _add_data($self, $data, $temp);
+    _add_data($self, $data, $tempdir);
     $dir = _insertion_directory($self);
     while (1) {
-        $new = $dir . "/" . _name($self->{rndhex});
-        $path = $self->{path} . "/" . $new;
-        rename($temp, $path) and return($new);
-        dief("cannot rename(%s, %s): %s", $temp, $path, $!)
+        $new = $dir."/"._name($self->{rndhex});
+        $path = $self->{path}."/".$new;
+        rename($tempdir, $path) and return($new);
+        dief("cannot rename(%s, %s): %s", $tempdir, $path, $!)
             unless $! == ENOTEMPTY or $! == EEXIST;
         # RACE: the target directory was already present...
     }
@@ -608,12 +650,12 @@ sub _volatile ($) {
 
     foreach my $name (_special_getdir($self->{path} .
                                       "/" . TEMPORARY_DIRECTORY)) {
-        push(@list, TEMPORARY_DIRECTORY . "/" . $1)
+        push(@list, TEMPORARY_DIRECTORY."/".$1)
             if $name =~ /^($_ElementRegexp)$/o; # untaint
     }
     foreach my $name (_special_getdir($self->{path} .
                                       "/" . OBSOLETE_DIRECTORY)) {
-        push(@list, OBSOLETE_DIRECTORY . "/" . $1)
+        push(@list, OBSOLETE_DIRECTORY."/".$1)
             if $name =~ /^($_ElementRegexp)$/o; # untaint
     }
     return(@list);
@@ -629,11 +671,11 @@ sub _destroy_dir ($) {
 
     foreach my $name (_special_getdir($dir)) {
         next if $name eq LOCKED_DIRECTORY;
-        $path = $dir . "/" . $name;
+        $path = $dir."/".$name;
         unlink($path) and next;
         dief("cannot unlink(%s): %s", $path, $!) unless $! == ENOENT;
     }
-    _special_rmdir($dir . "/" . LOCKED_DIRECTORY);
+    _special_rmdir($dir."/".LOCKED_DIRECTORY);
     _special_rmdir($dir);
 }
 
@@ -651,8 +693,8 @@ sub purge : method {
     my(@list, $path, $subdirs, $oldtime, $locked);
 
     # check options
-    $option{maxtemp} = 300 unless defined($option{maxtemp});
-    $option{maxlock} = 600 unless defined($option{maxlock});
+    $option{maxtemp} = $self->{maxtemp} unless defined($option{maxtemp});
+    $option{maxlock} = $self->{maxtemp} unless defined($option{maxlock});
     foreach my $name (keys(%option)) {
         dief("unexpected option: %s", $name)
             unless $name =~ /^(maxtemp|maxlock)$/;
@@ -669,7 +711,7 @@ sub purge : method {
         @list = sort(@list);
         pop(@list);
         foreach my $name (@list) {
-            $path = $self->{path} . "/" . $name;
+            $path = $self->{path}."/".$name;
             $subdirs = $_SubDirs->($path);
             next if $subdirs or not defined($subdirs);
             _special_rmdir($path);
@@ -679,7 +721,7 @@ sub purge : method {
     if ($option{maxtemp}) {
         $oldtime = time() - $option{maxtemp};
         foreach my $name (_volatile($self)) {
-            $path = $self->{path} . "/" . $name;
+            $path = $self->{path}."/".$name;
             next unless _older($path, $oldtime);
             warnf("removing too old volatile element: %s", $name);
             _destroy_dir($path);
@@ -758,22 +800,21 @@ Directory::Queue::Normal - object oriented interface to a normal directory based
 
 =head1 DESCRIPTION
 
-The goal of this module is to offer a "normal" (as opposed to
-"simple") queue system using the underlying filesystem for storage,
-security and to prevent race conditions via atomic operations.
+The goal of this module is to offer a "normal" (as opposed to "simple") queue
+system using the underlying filesystem for storage, security and to prevent
+race conditions via atomic operations.
 
-It allows arbitrary data to be stored (see the L</SCHEMA> section for
-more information) but it has a significant disk space and speed
-overhead.
+It allows arbitrary data to be stored (see the L</SCHEMA> section for more
+information) but it has a significant disk space and speed overhead.
 
-Please refer to L<Directory::Queue> for general information about
-directory queues.
+Please refer to L<Directory::Queue> for general information about directory
+queues.
 
 =head1 CONSTRUCTOR
 
-The new() method can be used to create a Directory::Queue::Normal
-object that will later be used to interact with the queue. The
-following attributes are supported:
+The new() method can be used to create a Directory::Queue::Normal object that
+will later be used to interact with the queue. The following attributes are
+supported:
 
 =over
 
@@ -783,20 +824,29 @@ the queue toplevel directory (mandatory)
 
 =item rndhex
 
-the "random" hexadecimal digit to use in element names (aka I<R>) as a
-number between 0 and 15
-(default: randomly generated)
+the "random" hexadecimal digit to use in element names (aka I<R>) as a number
+between 0 and 15 (default: randomly generated)
 
 =item umask
 
-the umask to use when creating files and directories
-(default: use the running process' umask)
+the umask to use when creating files and directories (default: use the running
+process' umask)
 
 =item maxelts
 
 the maximum number of elements that an intermediate directory can hold
 (default: 16,000)
 
+=item maxlock
+
+default maximum time for a locked element (in seconds, default 600)
+as used by the purge() method
+
+=item maxtemp
+
+default maximum time for a temporary element (in seconds, default 300)
+as used by the purge() method
+
 =item schema
 
 the schema defining how to interpret user supplied data
@@ -806,14 +856,14 @@ the schema defining how to interpret user supplied data
 
 =head1 SCHEMA
 
-The schema defines how user supplied data is stored in the queue. It
-is only required by the add() and get() methods.
+The schema defines how user supplied data is stored in the queue. It is only
+required by the add() and get() methods.
 
 The schema must be a reference to a hash containing key/value pairs.
 
-The key must contain only alphanumerical characters. It identifies the
-piece of data and will be used as file name when storing the data
-inside the element directory.
+The key must contain only alphanumerical characters. It identifies the piece
+of data and will be used as file name when storing the data inside the element
+directory.
 
 The value represents the type of the given piece of data. It can be:
 
@@ -821,29 +871,28 @@ The value represents the type of the given piece of data. It can be:
 
 =item binary
 
-the data is a binary string (i.e. a sequence of bytes), it will be
-stored directly in a plain file with no further encoding
+the data is a binary string (i.e. a sequence of bytes), it will be stored
+directly in a plain file with no further encoding
 
 =item string
 
-the data is a text string (i.e. a sequence of characters), it will be
-UTF-8 encoded before being stored in a file
+the data is a text string (i.e. a sequence of characters), it will be UTF-8
+encoded before being stored in a file
 
 =item table
 
-the data is a reference to a hash of text strings, it will be
-serialized and UTF-8 encoded before being stored in a file
+the data is a reference to a hash of text strings, it will be serialized and
+UTF-8 encoded before being stored in a file
 
 =back
 
-By default, all pieces of data are mandatory. If you append a question
-mark to the type, this piece of data will be marked as optional. See
-the comments in the L</SYNOPSIS> section for an example.
+By default, all pieces of data are mandatory. If you append a question mark to
+the type, this piece of data will be marked as optional. See the comments in
+the L</SYNOPSIS> section for an example.
 
-By default, string or binary data is used directly. If you append an
-asterisk to the type, the data that you add or get will be by
-reference. This can be useful to avoid string copies of large amounts
-of data.
+By default, string or binary data is used directly. If you append an asterisk
+to the type, the data that you add or get will be by reference. This can be
+useful to avoid string copies of large amounts of data.
 
 =head1 METHODS
 
@@ -857,8 +906,8 @@ return a new Directory::Queue::Normal object (class method)
 
 =item copy()
 
-return a copy of the object; this can be useful to have independent
-iterators on the same queue
+return a copy of the object; this can be useful to have independent iterators
+on the same queue
 
 =item path()
 
@@ -884,27 +933,27 @@ return an empty string if there is no next element
 
 =item add(DATA)
 
-add the given data (a hash or hash reference) to the queue and return
-the corresponding element name; the schema must be known and the data
-must conform to it
+add the given data (a hash or hash reference) to the queue and return the
+corresponding element name; the schema must be known and the data must conform
+to it
 
 =item lock(ELEMENT[, PERMISSIVE])
 
 attempt to lock the given element and return true on success; if the
-PERMISSIVE option is true (which is the default), it is not a fatal
-error if the element cannot be locked and false is returned
+PERMISSIVE option is true (which is the default), it is not a fatal error if
+the element cannot be locked and false is returned
 
 =item unlock(ELEMENT[, PERMISSIVE])
 
 attempt to unlock the given element and return true on success; if the
-PERMISSIVE option is true (which is I<not> the default), it is not a
-fatal error if the element cannot be unlocked and false is returned
+PERMISSIVE option is true (which is I<not> the default), it is not a fatal
+error if the element cannot be unlocked and false is returned
 
 =item touch(ELEMENT)
 
 update the access and modification times on the element's directory to
-indicate that it is still being used; this is useful for elements that
-are locked for long periods of time (see the purge() method)
+indicate that it is still being used; this is useful for elements that are
+locked for long periods of time (see the purge() method)
 
 =item remove(ELEMENT)
 
@@ -913,28 +962,26 @@ remove the given element (which must be locked) from the queue
 =item get(ELEMENT)
 
 get the data from the given element (which must be locked) and return
-basically the same hash as what add() got (in list context, the hash
-is returned directly while in scalar context, the hash reference is
-returned instead); the schema must be knownand the data must conform
-to it
+basically the same hash as what add() got (in list context, the hash is
+returned directly while in scalar context, the hash reference is returned
+instead); the schema must be knownand the data must conform to it
 
 =item purge([OPTIONS])
 
-purge the queue by removing unused intermediate directories, removing
-too old temporary elements and unlocking too old locked elements (aka
-staled locks); note: this can take a long time on queues with many
-elements; OPTIONS can be:
+purge the queue by removing unused intermediate directories, removing too old
+temporary elements and unlocking too old locked elements (aka staled locks);
+note: this can take a long time on queues with many elements; OPTIONS can be:
 
 =over
 
 =item maxtemp
 
-maximum time for a temporary element (in seconds, default 300);
+maximum time for a temporary element (in seconds);
 if set to 0, temporary elements will not be removed
 
 =item maxlock
 
-maximum time for a locked element (in seconds, default 600);
+maximum time for a locked element (in seconds);
 if set to 0, locked elements will not be unlocked
 
 =back
@@ -943,9 +990,9 @@ if set to 0, locked elements will not be unlocked
 
 =head1 DIRECTORY STRUCTURE
 
-All the directories holding the elements and all the files holding the
-data pieces are located under the queue toplevel directory. This
-directory can contain:
+All the directories holding the elements and all the files holding the data
+pieces are located under the queue toplevel directory. This directory can
+contain:
 
 =over
 
@@ -959,13 +1006,13 @@ the directory holding obsolete elements, i.e. the elements being removed
 
 =item I<NNNNNNNN>
 
-an intermediate directory holding elements; I<NNNNNNNN> is an 8-digits
-long hexadecimal number
+an intermediate directory holding elements; I<NNNNNNNN> is an 8-digits long
+hexadecimal number
 
 =back
 
-In any of the above directories, an element is stored as a single
-directory with a 14-digits long hexadecimal name I<SSSSSSSSMMMMMR> where:
+In any of the above directories, an element is stored as a single directory
+with a 14-digits long hexadecimal name I<SSSSSSSSMMMMMR> where:
 
 =over
 
@@ -983,9 +1030,9 @@ is a random hexadecimal digit used to reduce name collisions
 
 =back
 
-Finally, inside an element directory, the different pieces of data are
-stored into different files, named according to the schema. A locked
-element contains in addition a directory named C<locked>.
+Finally, inside an element directory, the different pieces of data are stored
+into different files, named according to the schema. A locked element contains
+in addition a directory named C<locked>.
 
 =head1 SEE ALSO
 
@@ -995,4 +1042,4 @@ L<Directory::Queue>.
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright (C) CERN 2010-2013
+Copyright (C) CERN 2010-2018
