@@ -13,15 +13,16 @@
 package Directory::Queue::Simple;
 use strict;
 use warnings;
-our $VERSION  = "1.7_2";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.16 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.10";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.19 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
 #
 
-use Directory::Queue qw(_name SYSBUFSIZE /Regexp/ /file/ /special/);
+use Directory::Queue qw(_create _name _touch SYSBUFSIZE /Regexp/ /special/);
 use No::Worries::Die qw(dief);
+use No::Worries::File qw(file_read file_write);
 use No::Worries::Stat qw(ST_MTIME);
 use No::Worries::Warn qw(warnf);
 use POSIX qw(:errno_h);
@@ -52,7 +53,7 @@ sub new : method {
 
     # default object
     $self = __PACKAGE__->SUPER::_new(%option);
-    foreach my $name (qw(path umask)) {
+    foreach my $name (qw(path maxlock maxtemp rndhex umask)) {
         delete($option{$name});
     }
     # check granularity
@@ -86,27 +87,19 @@ sub _add_dir ($) {
 
 sub _add_data ($$) {
     my($self, $dataref) = @_;
-    my($dir, $name, $tmp, $fh, $length, $offset, $done);
+    my($dir, $name, $path, $fh);
 
     $dir = _add_dir($self);
     while (1) {
         $name = _name($self->{rndhex});
-        $tmp = $self->{path} . "/" . $dir . "/" . $name . TEMPORARY_SUFFIX;
-        $fh = _file_create($tmp, $self->{umask});
+        $path = $self->{path}."/".$dir."/".$name . TEMPORARY_SUFFIX;
+        $fh = _create($path, $self->{umask});
         last if $fh;
-        _special_mkdir($self->{path} . "/" . $dir, $self->{umask})
+        _special_mkdir($self->{path}."/".$dir, $self->{umask})
             if $! == ENOENT;
     }
-    $length = length(${ $dataref });
-    $offset = 0;
-    while ($length) {
-        $done = syswrite($fh, ${ $dataref }, SYSBUFSIZE, $offset);
-        dief("cannot syswrite(%s): %s", $tmp, $!) unless defined($done);
-        $length -= $done;
-        $offset += $done;
-    }
-    close($fh) or dief("cannot close(%s): %s", $tmp, $!);
-    return($dir, $tmp);
+    file_write($path, handle => $fh, data => $dataref);
+    return($dir, $path);
 }
 
 sub _add_path ($$$) {
@@ -115,11 +108,11 @@ sub _add_path ($$$) {
 
     while (1) {
         $name = _name($self->{rndhex});
-        $new = $self->{path} . "/" . $dir . "/" . $name;
+        $new = $self->{path}."/".$dir."/".$name;
         # N.B. we use link() + unlink() to make sure $new is never overwritten
         if (link($tmp, $new)) {
             unlink($tmp) or dief("cannot unlink(%s): %s", $tmp, $!);
-            return($dir . "/" . $name);
+            return($dir."/".$name);
         }
         dief("cannot link(%s, %s): %s", $tmp, $new, $!) unless $! == EEXIST;
     }
@@ -150,7 +143,7 @@ sub add_path : method {
     my($dir);
 
     $dir = _add_dir($self);
-    _special_mkdir($self->{path} . "/" . $dir, $self->{umask});
+    _special_mkdir($self->{path}."/".$dir, $self->{umask});
     return(_add_path($self, $path, $dir));
 }
 
@@ -161,19 +154,20 @@ sub add_path : method {
 sub get : method {
     my($self, $name) = @_;
 
-    return(${ _file_read_bin($self->{path} . "/" . $name . LOCKED_SUFFIX) });
+    return(file_read($self->{path}."/".$name . LOCKED_SUFFIX));
 }
 
 sub get_ref : method {
     my($self, $name) = @_;
+    my($data);
 
-    return(_file_read_bin($self->{path} . "/" . $name . LOCKED_SUFFIX));
+    return(file_read($self->{path}."/".$name . LOCKED_SUFFIX, data => \$data));
 }
 
 sub get_path : method {
     my($self, $name) = @_;
 
-    return($self->{path} . "/" . $name . LOCKED_SUFFIX);
+    return($self->{path}."/".$name . LOCKED_SUFFIX);
 }
 
 #
@@ -187,7 +181,7 @@ sub lock : method {  ## no critic 'ProhibitBuiltinHomonyms'
     my($path, $lock, $time, $ignored);
 
     $permissive = 1 unless defined($permissive);
-    $path = $self->{path} . "/" . $name;
+    $path = $self->{path}."/".$name;
     $lock = $path . LOCKED_SUFFIX;
     unless (link($path, $lock)) {
         return(0) if $permissive and ($! == EEXIST or $! == ENOENT);
@@ -222,13 +216,23 @@ sub unlock : method {
     my($path, $lock);
 
     $permissive = 0 unless defined($permissive);
-    $path = $self->{path} . "/" . $name;
+    $path = $self->{path}."/".$name;
     $lock = $path . LOCKED_SUFFIX;
     return(1) if unlink($lock);
     return(0) if $permissive and $! == ENOENT;
     dief("cannot unlink(%s): %s", $lock, $!);
 }
 
+#
+# touch an element to indicate that it is still being used
+#
+
+sub touch : method {
+    my($self, $element) = @_;
+
+    _touch($self->{"path"}."/".$element);
+}
+
 #
 # remove a locked element from the queue
 #
@@ -237,7 +241,7 @@ sub remove : method {
     my($self, $name) = @_;
     my($path, $lock);
 
-    $path = $self->{path} . "/" . $name;
+    $path = $self->{path}."/".$name;
     $lock = $path . LOCKED_SUFFIX;
     unlink($path) or dief("cannot unlink(%s): %s", $path, $!);
     unlink($lock) or dief("cannot unlink(%s): %s", $lock, $!);
@@ -259,7 +263,7 @@ sub count : method {
     # count the elements inside
     foreach my $name (@list) {
         $count += grep(/^(?:$_ElementRegexp)$/o,
-                       _special_getdir($self->{path} . "/" . $name));
+                       _special_getdir($self->{path}."/".$name));
     }
     # that's all
     return($count);
@@ -274,7 +278,7 @@ sub _purge_dir ($$$) {
     my($path, @stat);
 
     foreach my $name (grep(/\./, _special_getdir($dir))) {
-        $path = $dir . "/" . $name;
+        $path = $dir."/".$name;
         @stat = stat($path);
         unless (@stat) {
             dief("cannot stat(%s): %s", $path, $!) unless $! == ENOENT;
@@ -299,8 +303,8 @@ sub purge : method {
     my(@list, $path, $oldtemp, $oldlock);
 
     # check options
-    $option{maxtemp} = 300 unless defined($option{maxtemp});
-    $option{maxlock} = 600 unless defined($option{maxlock});
+    $option{maxtemp} = $self->{maxtemp} unless defined($option{maxtemp});
+    $option{maxlock} = $self->{maxtemp} unless defined($option{maxlock});
     foreach my $name (keys(%option)) {
         dief("unexpected option: %s", $name)
             unless $name =~ /^(maxtemp|maxlock)$/;
@@ -318,7 +322,7 @@ sub purge : method {
     $oldlock = time() - $option{maxlock} if $option{maxlock};
     if ($oldtemp or $oldlock) {
         foreach my $name (@list) {
-            _purge_dir($self->{path} . "/" . $name, $oldtemp, $oldlock);
+            _purge_dir($self->{path}."/".$name, $oldtemp, $oldlock);
         }
     }
     # try to purge all but the last intermediate directory
@@ -326,7 +330,7 @@ sub purge : method {
         @list = sort(@list);
         pop(@list);
         foreach my $name (@list) {
-            $path = $self->{path} . "/" . $name;
+            $path = $self->{path}."/".$name;
             _special_rmdir($path) unless _special_getdir($path);
         }
     }
@@ -369,20 +373,20 @@ Directory::Queue::Simple - object oriented interface to a simple directory based
 
 =head1 DESCRIPTION
 
-The goal of this module is to offer a "simple" (as opposed to
-"normal") queue system using the underlying filesystem for storage,
-security and to prevent race conditions via atomic operations.
+The goal of this module is to offer a "simple" (as opposed to "normal") queue
+system using the underlying filesystem for storage, security and to prevent
+race conditions via atomic operations.
 
 It only allows binary strings to be stored but it is fast and small.
 
-Please refer to L<Directory::Queue> for general information about
-directory queues.
+Please refer to L<Directory::Queue> for general information about directory
+queues.
 
 =head1 CONSTRUCTOR
 
-The new() method can be used to create a Directory::Queue::Simple
-object that will later be used to interact with the queue. The
-following attributes are supported:
+The new() method can be used to create a Directory::Queue::Simple object that
+will later be used to interact with the queue. The following attributes are
+supported:
 
 =over
 
@@ -392,14 +396,23 @@ the queue toplevel directory (mandatory)
 
 =item rndhex
 
-the "random" hexadecimal digit to use in element names (aka I<R>) as a
-number between 0 and 15
-(default: randomly generated)
+the "random" hexadecimal digit to use in element names (aka I<R>) as a number
+between 0 and 15 (default: randomly generated)
 
 =item umask
 
-the umask to use when creating files and directories
-(default: use the running process' umask)
+the umask to use when creating files and directories (default: use the running
+process' umask)
+
+=item maxlock
+
+default maximum time for a locked element (in seconds, default 600)
+as used by the purge() method
+
+=item maxtemp
+
+default maximum time for a temporary element (in seconds, default 300)
+as used by the purge() method
 
 =item granularity
 
@@ -420,8 +433,8 @@ return a new Directory::Queue::Simple object (class method)
 
 =item copy()
 
-return a copy of the object; this can be useful to have independent
-iterators on the same queue
+return a copy of the object; this can be useful to have independent iterators
+on the same queue
 
 =item path()
 
@@ -447,38 +460,38 @@ return an empty string if there is no next element
 
 =item add(DATA)
 
-add the given data (a binary string) to the queue and return the
-corresponding element name
+add the given data (a binary string) to the queue and return the corresponding
+element name
 
 =item add_ref(REF)
 
-add the given data (a reference to a binary string) to the queue and
-return the corresponding element name, this can avoid string copies
-with large strings
+add the given data (a reference to a binary string) to the queue and return
+the corresponding element name, this can avoid string copies with large
+strings
 
 =item add_path(PATH)
 
-add the given file (identified by its path) to the queue and return
-the corresponding element name, the file must be on the same
-filesystem and will be moved to the queue
+add the given file (identified by its path) to the queue and return the
+corresponding element name, the file must be on the same filesystem and will
+be moved to the queue
 
 =item lock(ELEMENT[, PERMISSIVE])
 
 attempt to lock the given element and return true on success; if the
-PERMISSIVE option is true (which is the default), it is not a fatal
-error if the element cannot be locked and false is returned
+PERMISSIVE option is true (which is the default), it is not a fatal error if
+the element cannot be locked and false is returned
 
 =item unlock(ELEMENT[, PERMISSIVE])
 
 attempt to unlock the given element and return true on success; if the
-PERMISSIVE option is true (which is I<not> the default), it is not a
-fatal error if the element cannot be unlocked and false is returned
+PERMISSIVE option is true (which is I<not> the default), it is not a fatal
+error if the element cannot be unlocked and false is returned
 
 =item touch(ELEMENT)
 
-update the access and modification times on the element's file to
-indicate that it is still being used; this is useful for elements that
-are locked for long periods of time (see the purge() method)
+update the access and modification times on the element's file to indicate
+that it is still being used; this is useful for elements that are locked for
+long periods of time (see the purge() method)
 
 =item remove(ELEMENT)
 
@@ -486,38 +499,35 @@ remove the given element (which must be locked) from the queue
 
 =item get(ELEMENT)
 
-get the data from the given element (which must be locked) and return
-a binary string
+get the data from the given element (which must be locked) and return a binary
+string
 
 =item get_ref(ELEMENT)
 
-get the data from the given element (which must be locked) and return
-a reference to a binary string, this can avoid string copies with
-large strings
+get the data from the given element (which must be locked) and return a
+reference to a binary string, this can avoid string copies with large strings
 
 =item get_path(ELEMENT)
 
-get the file path of the given element (which must be locked), this
-file can be read but not removed, you must use the remove() method for
-this
+get the file path of the given element (which must be locked), this file can
+be read but not removed, you must use the remove() method for this
 
 =item purge([OPTIONS])
 
-purge the queue by removing unused intermediate directories, removing
-too old temporary elements and unlocking too old locked elements (aka
-staled locks); note: this can take a long time on queues with many
-elements; OPTIONS can be:
+purge the queue by removing unused intermediate directories, removing too old
+temporary elements and unlocking too old locked elements (aka staled locks);
+note: this can take a long time on queues with many elements; OPTIONS can be:
 
 =over
 
 =item maxtemp
 
-maximum time for a temporary element (in seconds, default 300);
+maximum time for a temporary element (in seconds);
 if set to 0, temporary elements will not be removed
 
 =item maxlock
 
-maximum time for a locked element (in seconds, default 600);
+maximum time for a locked element (in seconds);
 if set to 0, locked elements will not be unlocked
 
 =back
@@ -526,24 +536,23 @@ if set to 0, locked elements will not be unlocked
 
 =head1 DIRECTORY STRUCTURE
 
-The toplevel directory contains intermediate directories that contain
-the stored elements, each of them in a file.
+The toplevel directory contains intermediate directories that contain the
+stored elements, each of them in a file.
 
 The names of the intermediate directories are time based: the element
-insertion time is used to create a 8-digits long hexadecimal number.
-The granularity (see the new() method) is used to limit the number of
-new directories. For instance, with a granularity of 60 (the default),
-new directories will be created at most once per minute.
+insertion time is used to create a 8-digits long hexadecimal number. The
+granularity (see the new() method) is used to limit the number of new
+directories. For instance, with a granularity of 60 (the default), new
+directories will be created at most once per minute.
 
-Since there is usually a filesystem limit in the number of directories
-a directory can hold, there is a trade-off to be made. If you want to
-support many added elements per second, you should use a low
-granularity to keep small directories. However, in this case, you will
-create many directories and this will limit the total number of
-elements you can store.
+Since there is usually a filesystem limit in the number of directories a
+directory can hold, there is a trade-off to be made. If you want to support
+many added elements per second, you should use a low granularity to keep small
+directories. However, in this case, you will create many directories and this
+will limit the total number of elements you can store.
 
-The elements themselves are stored in files (one per element) with a
-14-digits long hexadecimal name I<SSSSSSSSMMMMMR> where:
+The elements themselves are stored in files (one per element) with a 14-digits
+long hexadecimal name I<SSSSSSSSMMMMMR> where:
 
 =over
 
@@ -561,11 +570,10 @@ is a random hexadecimal digit used to reduce name collisions
 
 =back
 
-A temporary element (being added to the queue) will have a C<.tmp>
-suffix.
+A temporary element (being added to the queue) will have a C<.tmp> suffix.
 
-A locked element will have a hard link with the same name and the
-C<.lck> suffix.
+A locked element will have a hard link with the same name and the C<.lck>
+suffix.
 
 =head1 SEE ALSO
 
@@ -575,4 +583,4 @@ L<Directory::Queue>.
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright (C) CERN 2010-2013
+Copyright (C) CERN 2010-2018
